from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


def __getattr__(name):
    raise AttributeError()


laz = LazyImporter()


with capture_imports(laz, auto_export=True):
    pass




from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports

from .func_submod import import_in_function

laz = LazyImporter()

with capture_imports(laz):
    import functools
    from . import import_target
    import_in_function()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import ducktools.lazyimporter as lazyimporter

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Ducktools: Lazy Importer'
copyright = '2023, David C Ellis'
author = 'David C Ellis'
release = lazyimporter.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx.ext.autodoc", "sphinx_rtd_theme"]

autodoc_typehints = "description"
autoclass_content = "init"

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import sys

import pytest

import ducktools.lazyimporter as lazyimporter

from ducktools.lazyimporter import (
    LazyImporter,
    ModuleImport,
    FromImport,
    MultiFromImport,
    TryExceptImport,
    TryExceptFromImport,
    TryFallbackImport,
    get_importer_state,
)


class TestDirectImports:
    def test_module_import(self):
        """
        Test Basic Module import occurs when expected
        """
        laz = LazyImporter(
            [
                ModuleImport("example_1"),
            ]
        )

        assert "example_1" not in sys.modules
        laz.example_1
        assert "example_1" in sys.modules

        # Check the import is the correct object
        import example_1  # noqa  # pyright: ignore

        assert example_1 is laz.example_1

    def test_module_import_asname(self):
        laz = LazyImporter([ModuleImport("example_1", asname="ex1")])

        import example_1  # noqa  # pyright: ignore

        assert example_1 is laz.ex1

    def test_from_import(self):
        """
        Test a basic from import from a module
        """
        laz = LazyImporter(
            [
                FromImport("example_2", "item", asname="i"),
            ]
        )

        assert "example_2" not in sys.modules
        laz.i
        assert "example_2" in sys.modules

        assert laz.i == "example"

        import example_2  # noqa  # pyright: ignore

        assert example_2.item is laz.i

    def test_imports_submod_asname(self):
        laz_sub = LazyImporter([ModuleImport("ex_mod.ex_submod", asname="ex_submod")])

        assert laz_sub.ex_submod.name == "ex_submod"

    def test_submod_from(self):
        """
        Test a from import from a submodule
        """
        laz = LazyImporter(
            [
                FromImport("ex_mod.ex_submod", "name"),
            ]
        )

        assert laz.name == "ex_submod"

    def test_submod_multifrom(self):
        """
        Test a basic multi from import
        """
        laz = LazyImporter(
            [
                MultiFromImport("ex_mod.ex_submod", ["name", ("name2", "othername")]),
            ]
        )

        assert laz.name == "ex_submod"
        assert laz.othername == "ex_submod2"

    def test_try_except_import(self):
        """
        Test a basic try/except import
        """
        # When the first import fails
        laz = LazyImporter(
            [
                TryExceptImport("module_does_not_exist", "ex_mod", "ex_mod"),
            ]
        )

        assert laz.ex_mod.name == "ex_mod"

        # When the first import succeeds
        laz2 = LazyImporter(
            [
                TryExceptImport("ex_mod", "ex_othermod", "ex_mod"),
            ]
        )

        assert laz2.ex_mod.name == "ex_mod"

    def test_try_except_submod_import(self):
        """
        Test a try/except import with submodules
        """
        laz = LazyImporter(
            [
                TryExceptImport(
                    "module_does_not_exist", "ex_mod.ex_submod", "ex_submod"
                ),
            ]
        )

        assert laz.ex_submod.name == "ex_submod"

    def test_try_except_from_import(self):
        laz = LazyImporter(
            [TryExceptFromImport("ex_mod", "name", "ex_othermod", "name", "name")]
        )

        assert laz.name == "ex_mod"

        laz = LazyImporter(
            [
                TryExceptFromImport(
                    "module_does_not_exist", "name", "ex_mod.ex_submod", "name", "name"
                )
            ]
        )

        assert laz.name == "ex_submod"

    def test_try_fallback_import(self):
        # noinspection PyUnresolvedReferences
        import ex_mod
        test_obj = object()

        laz = LazyImporter(
            [TryFallbackImport("ex_mod", test_obj)]
        )

        assert laz.ex_mod is ex_mod

        laz = LazyImporter(
            [TryFallbackImport("ex_mod", test_obj, "module_name")]
        )

        assert laz.module_name is ex_mod

        laz = LazyImporter(
            [TryFallbackImport("module_does_not_exist", test_obj)]
        )

        assert laz.module_does_not_exist is test_obj

        laz = LazyImporter(
            [TryFallbackImport("module_does_not_exist", test_obj, "module_name")]
        )

        assert laz.module_name is test_obj


class TestRelativeImports:
    def test_relative_import(self):
        import example_modules.lazy_submod_ex as lse

        laz = lse.lazy_submod_from_import()
        assert laz.name == "ex_submod"

        laz = lse.lazy_submod_multi_from_import()
        assert laz.name == "ex_submod"
        assert laz.othername == "ex_submod2"

    def test_submod_relative_import(self):
        from example_modules.ex_othermod import laz

        assert laz.submod_name == "ex_submod"


class TestEager:
    def test_eager_process(self):
        laz = LazyImporter([ModuleImport("functools")], eager_process=False)

        assert "_importers" not in vars(laz)

        laz = LazyImporter([ModuleImport("functools")], eager_process=True)

        assert "_importers" in vars(laz)

    def test_eager_import(self):
        laz = LazyImporter([ModuleImport("functools")], eager_import=False)

        assert "functools" not in vars(laz)

        laz = LazyImporter([ModuleImport("functools")], eager_import=True)

        assert "functools" in vars(laz)

    def test_eager_process_glob(self):
        initial_state = lazyimporter.EAGER_PROCESS

        lazyimporter.EAGER_PROCESS = False

        # EAGER_PROCESS = False and no value - should lazily process
        laz = LazyImporter([ModuleImport("functools")])
        assert "_importers" not in vars(laz)

        # EAGER_PROCESS = False and eager_process = False - should lazily process
        laz = LazyImporter([ModuleImport("functools")], eager_process=False)
        assert "_importers" not in vars(laz)

        # EAGER_PROCESS = False and eager_process = True - should eagerly process
        laz = LazyImporter([ModuleImport("functools")], eager_process=True)
        assert "_importers" in vars(laz)

        lazyimporter.EAGER_PROCESS = True

        # EAGER_PROCESS = True and no value - should eagerly process
        laz = LazyImporter([ModuleImport("functools")])
        assert "_importers" in vars(laz)

        # EAGER_PROCESS = True and eager_process = False - should lazily process
        laz = LazyImporter([ModuleImport("functools")], eager_process=False)
        assert "_importers" not in vars(laz)

        # EAGER_PROCESS = True and eager_process = True - should eagerly process
        laz = LazyImporter([ModuleImport("functools")], eager_process=True)
        assert "_importers" in vars(laz)

        # Restore state
        lazyimporter.EAGER_PROCESS = initial_state

    def test_eager_import_glob(self):
        initial_state = lazyimporter.EAGER_IMPORT

        lazyimporter.EAGER_IMPORT = False

        # EAGER_IMPORT = False and no value - should lazily import
        laz = LazyImporter([ModuleImport("functools")])
        assert "functools" not in vars(laz)

        # EAGER_IMPORT = False and eager_import = False - should lazily import
        laz = LazyImporter([ModuleImport("functools")], eager_import=False)
        assert "functools" not in vars(laz)

        # EAGER_IMPORT = False and eager_import = True - should eagerly import
        laz = LazyImporter([ModuleImport("functools")], eager_import=True)
        assert "functools" in vars(laz)

        lazyimporter.EAGER_IMPORT = True

        # EAGER_IMPORT = True and no value - should eagerly import
        laz = LazyImporter([ModuleImport("functools")])
        assert "functools" in vars(laz)

        # EAGER_IMPORT = True and eager_import = False - should lazily import
        laz = LazyImporter([ModuleImport("functools")], eager_import=False)
        assert "functools" not in vars(laz)

        # EAGER_IMPORT = True and eager_import = True - should eagerly import
        laz = LazyImporter([ModuleImport("functools")], eager_import=True)
        assert "functools" in vars(laz)

        # Restore state
        lazyimporter.EAGER_IMPORT = initial_state



# README EXAMPLE CODE #

# NOTE: This is a simplified example using importlib.import_module
import importlib
from ducktools.lazyimporter import ImportBase, LazyImporter, FromImport


class IfElseImporter(ImportBase):
    def __init__(self, condition, module_name, else_module_name, asname):
        self.condition = condition
        self.module_name = module_name
        self.else_module_name = else_module_name
        self.asname = asname

        if not self.asname.isidentifier():
            raise ValueError(f"{self.asname} is not a valid python identifier.")

    def import_objects(self, globs=None):
        if globs is not None:
            package = globs.get('__name__')
        else:
            package = None

        if self.condition:
            mod = importlib.import_module(self.module_name, package)
        else:
            mod = importlib.import_module(self.else_module_name, package)

        return {self.asname: mod}


# Test for readme example code
def test_ifelse_importer():
    laz_if = LazyImporter(
        [
            IfElseImporter(
                condition=True,
                module_name="ex_mod",
                else_module_name="ex_othermod",
                asname="ex_mod",
            )
        ]
    )

    laz_else = LazyImporter(
        [
            IfElseImporter(
                condition=False,
                module_name="ex_mod",
                else_module_name="ex_othermod",
                asname="ex_mod",
            )
        ]
    )

    assert laz_if.ex_mod.name == "ex_mod"
    assert laz_else.ex_mod.name == "ex_othermod"


def test_doc_example():
    from dataclasses import dataclass

    laz = LazyImporter(
        [
            FromImport("dataclasses", "fields"),
            FromImport("json", "dumps"),
        ]
    )

    def _dataclass_default(dc):
        # In general is_dataclass should be used, but for this case
        # in order to demonstrate laziness it is not.
        if hasattr(dc, "__dataclass_fields__"):
            fields = laz.fields(dc)
            return {f.name: getattr(dc, f.name) for f in fields}
        raise TypeError("Object is not a Dataclass")

    def dumps(obj, **kwargs):
        default = kwargs.pop("default", None)
        if default:

            def new_default(o):
                try:
                    return default(o)
                except TypeError:
                    return _dataclass_default(o)

        else:
            new_default = _dataclass_default
        kwargs["default"] = new_default

        return laz.dumps(obj, **kwargs)

    @dataclass
    class TestEx:
        answer: int
        question: str

    in_dict = {
        "answer": 42,
        "question": "What do you get if you multiply 6 by 9?",
    }

    ex = TestEx(**in_dict)
    output = laz.dumps(in_dict)

    assert dumps(ex) == output



import pytest

from ducktools.lazyimporter import (
    ModuleImport,
    FromImport,
    MultiFromImport,
    TryExceptImport,
    TryExceptFromImport,
    LazyImporter,
)


def test_missing_import():
    laz = LazyImporter([ModuleImport("importlib")])

    with pytest.raises(AttributeError):
        _ = laz.missing_attribute


def test_missing_attribute():
    import ex_othermod

    assert ex_othermod.name == "ex_othermod"

    with pytest.raises(AttributeError) as e:
        _ = ex_othermod.invalid

    assert e.match("module 'ex_othermod' has no attribute 'invalid'")


def test_invalid_input():
    with pytest.raises(TypeError) as e:
        laz = LazyImporter(["importlib"], eager_process=True)

    assert e.match(
        "'importlib' is not an instance of "
        "ModuleImport, FromImport, MultiFromImport or TryExceptImport"
    )


def test_invalid_relative_import():
    with pytest.raises(ValueError) as e:
        _ = ModuleImport(".relative_module")

    assert e.match("Relative import '.relative_module' requires an assigned name.")


def test_invalid_submodule_import():
    with pytest.raises(ValueError) as e:
        _ = ModuleImport("importlib.util")

    assert e.match("Submodule import 'importlib.util' requires an assigned name.")


class TestInvalidIdentifiers:
    def test_modimport_invalid(self):
        with pytest.raises(ValueError) as e:
            _ = ModuleImport("modname", "##invalid_identifier##")

        assert e.match(f"'##invalid_identifier##' is not a valid Python identifier.")

    def test_fromimport_invalid(self):
        with pytest.raises(ValueError) as e:
            _ = FromImport("modname", "attribute", "##invalid_identifier##")

        assert e.match(f"'##invalid_identifier##' is not a valid Python identifier.")

    def test_multifromimport_invalid(self):
        with pytest.raises(ValueError) as e:
            _ = MultiFromImport("modname", [("attribute", "##invalid_identifier##")])

        assert e.match(f"'##invalid_identifier##' is not a valid Python identifier.")

    def test_tryexceptimport_invalid(self):
        with pytest.raises(ValueError) as e:
            _ = TryExceptImport("modname", "altmod", "##invalid_identifier##")

        assert e.match(f"'##invalid_identifier##' is not a valid Python identifier.")

    def test_tryexceptfromimport_invalid(self):
        with pytest.raises(ValueError) as e:
            _ = TryExceptFromImport(
                "modname",
                "attribname",
                "altmod",
                "altattribute",
                "##invalid_identifier##",
            )
        assert e.match(f"'##invalid_identifier##' is not a valid Python identifier.")


class TestNameClash:
    def test_moduleimport_clash(self):
        with pytest.raises(ValueError) as e:
            importer = LazyImporter(
                [ModuleImport("collections"), ModuleImport("collections")],
                eager_process=True,
            )
        assert e.match("'collections' used for multiple imports.")

    def test_fromimport_clash(self):
        """
        Multiple FromImports with clashing 'asname' parameters
        """

        with pytest.raises(ValueError) as e:
            laz = LazyImporter(
                [
                    FromImport("collections", "namedtuple", "nt"),
                    FromImport("typing", "NamedTuple", "nt"),
                ],
                eager_process=True,
            )

        assert e.match("'nt' used for multiple imports.")

    def test_multifromimport_clash(self):
        """
        Multiple FromImports with clashing 'asname' parameters
        """

        with pytest.raises(ValueError) as e:
            laz = LazyImporter(
                [
                    MultiFromImport(
                        "collections", [("namedtuple", "nt"), ("defaultdict", "nt")]
                    ),
                ],
                eager_process=True,
            )

        assert e.match("'nt' used for multiple imports.")

    def test_mixedimport_clash(self):
        with pytest.raises(ValueError) as e:
            laz = LazyImporter(
                [
                    FromImport("mod1", "matching_mod_name"),
                    ModuleImport("matching_mod_name"),
                ],
                eager_process=True,
            )

        assert e.match("'matching_mod_name' used for multiple imports.")

    def test_reserved_name(self):
        with pytest.raises(ValueError) as e:
            laz = LazyImporter(
                [
                    FromImport("mod1", "objname", "_importers"),
                ],
                eager_process=True,
            )

        assert e.match("'_importers' clashes with a LazyImporter internal name.")

class TestNoGlobals:
    def test_relative_module_noglobals(self):
        """
        ModuleImport relative without globals where frame hacks are unavailable
        """
        laz = LazyImporter(
            [ModuleImport(".relative_module", asname="relative_module")],
            eager_process=False,
            eager_import=False,
        )

        laz._globals = None

        with pytest.raises(ValueError) as e:
            _ = laz._importers

        assert e.match(
            "Attempted to setup relative import without providing globals()."
        )

    def test_relative_from_noglobals(self):
        """
        FromImport relative without globals
        """
        laz = LazyImporter(
            [FromImport(".relative_module", "attribute")],
            eager_process=False,
            eager_import=False,
        )

        laz._globals = None

        with pytest.raises(ValueError) as e:
            _ = laz._importers

        assert e.match(
            "Attempted to setup relative import without providing globals()."
        )


class TestImportErrors:
    def test_module_import_nosubmod_asname(self):
        laz = LazyImporter(
            [
                ModuleImport("importlib.util.fakemod", asname="fakemod"),
            ]
        )

        with pytest.raises(ModuleNotFoundError) as e:
            _ = laz.fakemod

        assert e.match("No module named 'importlib.util.fakemod'")

    def test_tryexcept_import_nosubmod_asname(self):
        laz = LazyImporter(
            [
                TryExceptImport(
                    "importlib.util.fakemod1",
                    "importlib.util.fakemod",
                    asname="fakemod",
                ),
            ]
        )

        with pytest.raises(ModuleNotFoundError) as e:
            _ = laz.fakemod

        assert e.match("No module named 'importlib.util.fakemod'")



from collections.abc import Callable
from typing import Any
from typing_extensions import Self

from . import LazyImporter

class CaptureError(Exception):
    ...

class _ImportPlaceholder:
    capturer: capture_imports
    attrib_name: str
    placeholder_parent: _ImportPlaceholder

    def __init__(
        self,
        capturer: capture_imports,
        attrib_name: str | None = None,
        parent: _ImportPlaceholder | None = None
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __getattr__(self, item: str) -> _ImportPlaceholder: ...


class CapturedModuleImport:
    module_name: str
    placeholder: _ImportPlaceholder

    def __init__(self, module_name: str, placeholder: _ImportPlaceholder) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __repr__(self) -> str: ...
    @property
    def final_element(self) -> str: ...

class CapturedFromImport:
    module_name: str
    attrib_name: str
    placeholder: _ImportPlaceholder

    def __init__(self, module_name: str, attrib_name: str, placeholder: _ImportPlaceholder) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __repr__(self) -> str: ...


_import_signature = Callable[
    [str, dict[str, Any] | None, dict[str, Any] | None, tuple[str, ...] | None, int],
    Any,
]

class capture_imports:
    importer: LazyImporter
    auto_export: bool

    captured_imports: list[CapturedModuleImport | CapturedFromImport]
    import_func: _import_signature | None
    previous_import_func: _import_signature | None
    globs: dict[str, Any] | None

    def __init__(self, importer: LazyImporter, auto_export: bool = True) -> None: ...

    def _make_capturing_import(self) -> _import_signature: ...

    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...



from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


laz = LazyImporter()
with capture_imports(laz, auto_export=False):
    import importlib.util
    import importlib



import builtins
import sys

from . import ModuleImport, MultiFromImport, extend_imports, get_module_funcs


class CaptureError(Exception):
    """
    Error for capture specific details
    """


# A replaced __import__ will still populate module globals (but not sys.modules)
# After the fake imports are done, these will be used to remove the names from the module
# This intentionally does **not** define an '__eq__' or '__hash__'
# Two placeholders are only identical if they are the same object as otherwise it's
# not possible to trace imports back correctly.
class _ImportPlaceholder:
    __slots__ = ("attrib_name", "placeholder_parent", "capturer")

    def __init__(self, capturer, attrib_name=None, parent=None):
        self.capturer = capturer
        self.attrib_name = attrib_name
        self.placeholder_parent = parent

    def __repr__(self):  # pragma: nocover
        return (
            f"{self.__class__.__name__}("
            f"capturer={self.capturer!r}, "
            f"attrib_name={self.attrib_name!r}, "
            f"parent={self.placeholder_parent!r}"
            f")"
        )

    def __getattr__(self, item):
        # 'as' imports will be created from attribute access
        # store the parent so we can work backwards to the original import
        return _ImportPlaceholder(capturer=self.capturer, attrib_name=item, parent=self)


# Temporary importers needed for tracking back to assigned names
class CapturedModuleImport:
    __slots__ = ("module_name", "placeholder")

    def __init__(self, module_name, placeholder):
        self.module_name = module_name
        self.placeholder = placeholder

    def __repr__(self):  # pragma: nocover
        return (
            f"{self.__class__.__name__}("
            f"module_name={self.module_name!r}, "
            f"placeholder={self.placeholder!r}"
            f")"
        )

    @property
    def final_element(self):
        return self.module_name.split(".")[-1]


class CapturedFromImport:
    __slots__ = ("module_name", "attrib_name", "placeholder")

    def __init__(self, module_name, attrib_name, placeholder):
        self.module_name = module_name
        self.attrib_name = attrib_name
        self.placeholder = placeholder

    def __repr__(self):  # pragma: nocover
        return (
            f"{self.__class__.__name__}("
            f"module_name={self.module_name!r}, "
            f"attrib_name={self.attrib_name!r}, "
            f"placeholder={self.placeholder!r}"
            f")"
        )


class capture_imports:
    def __init__(self, importer, auto_export=True):
        """
        Capture import statements executed within a block using this as a context manager

        Any `import <module> as <name>` or `from <module> import <attribute> as <name>`
        calls will be intercepted and assigned as lazy imports to the importer provided.

        **This only works at module level!**

        :param importer: LazyImporter instance
        :param auto_export: generate __getattr__ and __dir__ functions and add them to globals
                            in order to make the attributes available to import on the module.
        """
        self.importer = importer
        self.auto_export = auto_export

        # Global and locals checks
        # This capture function only works at module level, so locals should be globals
        # And they should match the level of the lazy importer
        try:
            sys._getframe  # noqa
        except AttributeError:
            raise CaptureError("Import capture requires sys._getframe")

        self.globs = None
        self.captured_imports = []

        # Place to store current and previous __import__ functions
        self.import_func = None
        self.previous_import_func = None

    def _make_capturing_import(self):
        def _capturing_import(name, globals=None, locals=None, fromlist=(), level=0):
            # Something else tried to import - redirect to regular machinery
            if globals is not self.globs or globals != locals:
                return self.previous_import_func(name, globals, locals, fromlist, level)

            if fromlist and "*" in fromlist:
                raise CaptureError("Lazy importers cannot capture '*' imports.")

            # Make a unique placeholder object
            placeholder = _ImportPlaceholder(
                capturer=self,
                attrib_name=name.split(".")[0],
            )

            leading_dots = "." * level
            module_name = f"{leading_dots}{name}"

            captured_imports = self.captured_imports
            if fromlist:
                captured_imports.extend(
                    [
                        CapturedFromImport(
                            module_name=module_name,
                            attrib_name=obj_name,
                            placeholder=placeholder,
                        )
                        for obj_name in fromlist
                    ]
                )
            else:
                captured_imports.append(
                    CapturedModuleImport(
                        module_name=module_name,
                        placeholder=placeholder,
                    )
                )

            return placeholder

        return _capturing_import

    def __enter__(self):
        if self.previous_import_func or self.import_func:
            raise CaptureError("_CaptureContext is not reusable")

        # Only capture the frame when entering, this makes it possible to wrap
        # capture_imports.
        frame = sys._getframe(1)
        globs = frame.f_globals
        locs = frame.f_locals

        if globs is not locs:
            raise CaptureError("Import capture must be done at module level")

        if globs is not self.importer._globals:
            raise CaptureError("LazyImporter globals must match frame globals for capture_imports call")

        if self.auto_export and (globs.get("__getattr__") or globs.get("__dir__")):
            raise CaptureError(
                "auto_export is not supported if __getattr__ or __dir__ is already defined on the module."
            )

        self.globs = globs

        # Store the old import function, create the new one and replace it
        self.previous_import_func = builtins.__import__

        self.import_func = self._make_capturing_import()

        builtins.__import__ = self.import_func

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore import state
        if builtins.__import__ is not self.import_func:
            # If the importer has been changed while in use restore it anyway
            # But error out
            builtins.__import__ = self.previous_import_func
            raise CaptureError(
                "Importer was replaced while in import block. "
                "Restored to state when block entered."
            )

        builtins.__import__ = self.previous_import_func

        # Trace names that are now in globals back to the actual import

        # `captured_imports` contains a list of CapturedImport objects containing
        # the import module and attribute names alongside a placeholder that was returned
        # by the original import
        # The globals namespace now contains the final placeholders that were assigned.
        # These may not be the original placeholders returned as 'from' imports for example
        # will have accessed attributes from the placeholders.
        # The logic is to trace the placeholder back to the parent, and then match that
        # with the placeholders contained by the CapturedImports

        # First create a mapping from placeholder instances to attribute names, to importers
        placeholders = {}
        for importer in self.captured_imports:
            importer_placeholders = placeholders.get(importer.placeholder, {})
            if isinstance(importer, CapturedModuleImport):
                importer_placeholders[importer.final_element] = importer
            else:
                importer_placeholders[importer.attrib_name] = importer

            placeholders[importer.placeholder] = importer_placeholders

        final_imports = []  # List of ModuleImport and MultiFromImport
        from_imports = {}  # dict of module_name: [(attrib, asname), ..]

        # Create a set of imports - if there is an importer that is not used
        # After all are matched then raise an exception
        importer_set = set(self.captured_imports)

        # Copy as the original may be mutated.
        # `key` in this case will be the name the attribute was assigned
        for key, value in self.globs.copy().items():
            if isinstance(value, _ImportPlaceholder) and value.capturer is self:
                # Store the initial attribute name and seek upwards through
                # parent placeholders to try to find the original placeholder
                attrib_name = value.attrib_name
                while parent := value.placeholder_parent:
                    value = parent

                # Get the {attribute name: importer, ...} mappings
                importer_map = placeholders[value]

                if attrib_name:
                    # Retrieve the captured import statement from the mapping
                    try:
                        capture = importer_map[attrib_name]
                    except KeyError:
                        # Search the capture map to see if this is a submodule import
                        for cap_imp in importer_map.values():
                            if cap_imp.module_name.split(".")[0] == attrib_name:
                                asname = cap_imp.module_name.split(".")[-1]
                                raise CaptureError(
                                    f"Submodule import `import {cap_imp.module_name}` requires assigned name: "
                                    f"eg `import {cap_imp.module_name} as {asname}`"
                                )

                        # I don't know of a case where this can currently happen
                        # But this is still an error so raise here
                        raise  # pragma: nocover

                    # Remove used importers from the set
                    importer_set.discard(capture)

                    # Convert it to a regular ModuleImport or store it to make
                    # a MultiFromImport at the end.
                    if isinstance(capture, CapturedModuleImport):
                        importer = ModuleImport(
                            capture.module_name,
                            asname=key,
                        )
                        final_imports.append(importer)
                    else:
                        try:
                            pairs = from_imports[capture.module_name]
                        except KeyError:
                            pairs = []
                            from_imports[capture.module_name] = pairs

                        pairs.append((attrib_name, key))

                    del self.globs[key]

        # importer_set should be empty
        # If it was not then there were import statements that did not have
        # a matching name in globals.

        if importer_set:
            missing_module_imports = []
            missing_from_imports = []
            for imp in importer_set:
                if isinstance(imp, CapturedModuleImport):
                    missing_module_imports.append(repr(imp.module_name))
                else:
                    missing_from_imports.append(repr(imp.attrib_name))

            if missing_module_imports:
                err_list = ", ".join(missing_module_imports)
                raise CaptureError(
                    f"Unused module import(s) {err_list} not matched to a name in globals. "
                    f"Possible duplicates or unaliased submodule import before module import."
                )
            else:
                err_list = ", ".join(missing_from_imports)
                raise CaptureError(f"Imports for name(s) {err_list} found, but unused.")

        final_imports.extend([MultiFromImport(k, v) for k, v in from_imports.items()])

        # Add these imports to the importer
        extend_imports(self.importer, final_imports)

        # Export to globals
        if self.auto_export:
            module_name = self.globs["__name__"]
            getattr_func, dir_func = get_module_funcs(self.importer, module_name=module_name)
            self.globs["__getattr__"] = getattr_func
            self.globs["__dir__"] = dir_func



from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


def __dir__():
    return []


laz = LazyImporter()


with capture_imports(laz, auto_export=True):
    pass




from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


def import_check():
    import functools
    return functools


laz = LazyImporter()
with capture_imports(laz, auto_export=False):
    import collections
    inner_import = import_check()
    class InnerClass:
        import typing



import builtins
import importlib

from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


laz = LazyImporter()

with capture_imports(laz, auto_export=False):
    builtins.__import__ = importlib.__import__
